"""
Built-in extensions for snek.
"""
import argparse
import sys
import textwrap
from typing import Callable, Iterable, List, Optional, Type

from ..actions import Action, register, unregister
from ..exceptions import ErrorLoadingExtension
from ..identification import dasherize, deterministic_sort, underscore

if sys.version_info[:2] >= (3, 8):
    # TODO: Import directly (no need for conditional) when `python_requires = >= 3.8`
    from importlib.metadata import EntryPoint, entry_points  # pragma: no cover
else:
    from importlib_metadata import EntryPoint, entry_points  # pragma: no cover


ENTRYPOINT_GROUP = "snek.cli"

NO_LONGER_NEEDED = {"pyproject", "tox"}
"""Extensions that are no longer needed and are now part of snek itself"""

# TODO: NO_LONGER_SUPPORTED = {"no_pyproject"}


class Extension:
    """Base class for snek's extensions

    Args:
        name (str): How the extension should be named. Default: name of class
            By default, this value is used to create the activation flag in
            snek cli.

    See our docs on how to create extensions in:
    https://snek.org/en/latest/extensions.html

    Also check :obj:`~snek.actions`, :obj:`~snek.structure.Structure` and
    :obj:`~snek.operations.ScaffoldOpts` for more details.

    Note:
        Please name your class using a CamelCase version of the name you use in the
        setuptools entrypoint (alternatively you will need to overwrite the ``name``
        property to match the entrypoint name).
    """

    #: When ``True`` snek will store the extension in the snek's section of
    #: ``setup.cfg``. Useful for updates. Set to ``False`` if the extension should not
    #: be re-invoked on updates.
    persist = True

    def __init__(self, name: Optional[str] = None):
        self._name = name or underscore(self.__class__.__name__)

    @property
    def name(self):
        return self._name

    @property
    def flag(self) -> str:
        return f"--{dasherize(self.name)}"

    @property
    def help_text(self) -> str:
        if self.__doc__ is None:
            raise NotImplementedError("Please provide a help text for your extension")

        doc = textwrap.dedent(self.__doc__)
        return doc[0].lower() + doc[1:]

    def augment_cli(self, parser: argparse.ArgumentParser):
        """Augments the command-line interface parser.

        A command line argument ``--FLAG`` where FLAG=``self.name`` is added
        which appends ``self.activate`` to the list of extensions. As help
        text the docstring of the extension class is used.
        In most cases this method does not need to be overwritten.

        Args:
            parser: current parser object
        """
        parser.add_argument(
            self.flag,
            dest="extensions",
            action="append_const",
            const=self,
            help=self.help_text,
        )
        return self

    def activate(self, actions: List[Action]) -> List[Action]:
        """Activates the extension by registering its functionality

        Args:
            actions (List[Action]): list of action to perform

        Returns:
            List[Action]: updated list of actions
        """
        raise NotImplementedError(f"Extension {self.name} has no actions registered")

    #: Shortcut for :obj:`snek.actions.register`
    register = staticmethod(register)

    #: Shortcut for :obj:`snek.actions.unregister`
    unregister = staticmethod(unregister)

    def __call__(self, actions: List[Action]) -> List[Action]:
        """Just delegating to :obj:`self.activate`"""
        return self.activate(actions)


def include(*extensions: Extension) -> Type[argparse.Action]:
    """Create a custom :obj:`argparse.Action` that saves multiple extensions for
    activation.

    Args:
        *extensions: extension objects to be saved
    """

    class IncludeExtensions(argparse.Action):
        """Appends the given extensions to the extensions list."""

        def __call__(self, parser, namespace, values, option_string=None):
            ext_list = list(getattr(namespace, "extensions", []))
            namespace.extensions = ext_list + list(extensions)

    return IncludeExtensions


def store_with(*extensions: Extension) -> Type[argparse.Action]:
    """Create a custom :obj:`argparse.Action` that stores the value of the given option
    in addition to saving the extension for activation.

    Args:
        *extensions: extension objects to be saved for activation
    """

    class AddExtensionAndStore(include(*extensions)):  # type: ignore
        """\
        Consumes the values provided, but also appends the given extension
        to the extensions list.
        """

        def __call__(self, parser, namespace, values, option_string=None):
            super().__call__(parser, namespace, values, option_string)
            setattr(namespace, self.dest, values)

    return AddExtensionAndStore


def iterate_entry_points(group=ENTRYPOINT_GROUP) -> Iterable[EntryPoint]:
    """Produces a generator yielding an EntryPoint object for each extension registered
    via `setuptools`_ entry point mechanism.

    This method can be used in conjunction with :obj:`load_from_entry_point` to filter
    the extensions before actually loading them.


    .. _setuptools: https://setuptools.pypa.io/en/latest/userguide/entry_point.html
    """  # noqa
    entries = entry_points()
    if hasattr(entries, "select"):
        # The select method was introduced in importlib_metadata 3.9 (and Python 3.10)
        # and the previous dict interface was declared deprecated
        return entries.select(group=group)  # type: ignore
    else:
        # TODO: Once Python 3.10 becomes the oldest version supported, this fallback and
        #       conditional statement can be removed.
        return (extension for extension in entries.get(group, []))  # type: ignore


def load_from_entry_point(entry_point: EntryPoint) -> Extension:
    """Carefully load the extension, raising a meaningful message in case of errors"""
    try:
        return entry_point.load()(entry_point.name)
    except Exception as ex:
        raise ErrorLoadingExtension(entry_point=entry_point) from ex


def list_from_entry_points(
    group: str = ENTRYPOINT_GROUP,
    filtering: Callable[[EntryPoint], bool] = lambda _: True,
) -> List[Extension]:
    """Produces a list of extension objects for each extension registered
    via `setuptools`_ entry point mechanism.

    Args:
        group: name of the setuptools' entry_point group where extensions is being
            registered
        filtering: function returning a boolean deciding if the entry point should be
            loaded and included (or not) in the final list. A ``True`` return means the
            extension should be included.

    .. _setuptools: https://setuptools.pypa.io/en/latest/userguide/entry_point.html
    """  # noqa
    return deterministic_sort(
        load_from_entry_point(e) for e in iterate_entry_points(group) if filtering(e)
    )



from setuptools import setup

if __name__ == "__main__":
    try:
        setup(
            use_scm_version={
                "local_scheme": lambda version: "",
                "version_scheme": lambda version: version.tag.base_version
            }
        )
    except:  # noqa
        print(
            "\n\nAn error occurred while building the project, "
            "please ensure you have the most updated version of setuptools, "
            "setuptools_scm and wheel with:\n"
            "   pip install -U setuptools setuptools_scm wheel\n\n"
        )
        raise



from importlib import reload
from io import StringIO

import pytest

from snek import termui


@pytest.fixture(scope="module")
def after():
    # Reload termui after tests to ensure constants are calculated
    # with original logic (without mocks).
    yield
    reload(termui)


@pytest.fixture
def fake_tty(monkeypatch):
    # NOTE:
    # This fixture is a workaround for the limitations of pytest stdout/stderr
    # capture.
    # To be realistic we should test termui against real TTY devices,
    # and maybe the best way of doing it is just using the stdout/stderr
    # in the terminal.
    # Since pytest buffers the output, we need to rely on the `disabled`
    # contextmanager of the `capsys` fixture.
    # However, this currently doesn't work with pytest-xdist (see
    # https://github.com/pytest-dev/pytest/issues/1991).
    # So if we intend to run our tests in parallel, we have to gave up on it :(
    stream = StringIO()
    monkeypatch.setattr(stream, "isatty", lambda *_: True)

    return stream


def test_isatty_file(tmpfolder, orig_isatty):
    file = tmpfolder.join("file.txt").ensure().open()
    assert not termui.isatty(file)


def test_isatty_buffer(orig_isatty):
    assert not termui.isatty(StringIO())


def test_isatty_random_obj(orig_isatty):
    assert not termui.isatty([1, 2, 3])


def test_isatty_tty(fake_tty, orig_isatty):
    assert termui.isatty(fake_tty)


def test_support_with_curses_no_colorama(
    fake_tty, curses_mock, no_colorama_mock, orig_isatty
):
    reload(termui)  # ensure mocks side effects
    assert termui.SYSTEM_SUPPORTS_COLOR
    assert termui.supports_color(fake_tty)


def test_support_no_curses_with_colorama(
    fake_tty, no_curses_mock, colorama_mock, orig_isatty
):
    reload(termui)  # ensure mocks side effects
    assert termui.SYSTEM_SUPPORTS_COLOR
    assert termui.supports_color(fake_tty)


def test_support_with_curses_with_colorama(
    fake_tty, curses_mock, colorama_mock, orig_isatty
):
    reload(termui)  # ensure mocks side effects
    assert termui.SYSTEM_SUPPORTS_COLOR
    assert termui.supports_color(fake_tty)


def test_support_no_colorama_no_curses(
    fake_tty, no_curses_mock, no_colorama_mock, orig_isatty
):
    reload(termui)  # ensure mocks side effects
    assert not termui.SYSTEM_SUPPORTS_COLOR
    assert not termui.supports_color(fake_tty)


def test_decorate():
    # When styles are passed,
    text = termui.decorate("text", "red", "bold")
    # then they should be present in the response
    assert "\033[1m" in text  # bold
    assert "\033[31m" in text  # red
    assert "\033[0m" in text  # clear

    # When styles are not passed,
    text = termui.decorate("text")
    # then the text should not contain ansi codes
    assert "\033[1m" not in text  # bold
    assert "\033[31m" not in text  # red
    assert "\033[0m" not in text  # clear
    assert text == "text"



#!/usr/bin/env python
import sys
from pathlib import Path

import pytest

from snek.actions import get_default_options
from snek.api import create_project
from snek.cli import parse_args, run
from snek.exceptions import InvalidIdentifier
from snek.extensions.namespace import (
    Namespace,
    add_namespace,
    enforce_namespace_options,
    move_old_package,
    prepare_namespace,
)
from snek.log import logger


def test_prepare_namespace():
    namespaces = prepare_namespace("com")
    assert namespaces == ["com"]
    namespaces = prepare_namespace("com.blue_yonder")
    assert namespaces == ["com", "com.blue_yonder"]
    with pytest.raises(InvalidIdentifier):
        prepare_namespace("com.blue-yonder")


def test_add_namespace(isolated_logger, caplog):
    args = ["project", "-p", "package", "--namespace", "com.blue_yonder"]
    opts = parse_args(args)
    opts["ns_list"] = prepare_namespace(opts["namespace"])
    struct = {"src": {"package": {"file1": "Content"}}}

    ns_struct, _ = add_namespace(struct, opts)

    ns_pkg_struct = ns_struct["src"]
    assert "project" not in set(ns_pkg_struct.keys())
    assert "package" not in set(ns_pkg_struct.keys())
    assert {"com"} == set(ns_pkg_struct.keys())
    modules = set(ns_pkg_struct["com"].keys())
    assert {"blue_yonder", "__init__.py"} == modules
    submodules = set(ns_pkg_struct["com"]["blue_yonder"].keys())
    assert "package" in submodules

    # warnings should not be logged
    log = caplog.text
    unexpected_log = ("empty namespace", "valid string", "the `--namespace` option")
    for text in unexpected_log:
        assert text not in log


def test_add_namespace_empty(isolated_logger, caplog):
    # When trying to add a namespace without actually providing a namespace string
    add_namespace({}, {})

    # then something should be logged,
    log = caplog.text
    expected_log = ("empty namespace", "valid string", "the `--namespace` option")
    for text in expected_log:
        assert text in log


def test_create_project_with_namespace(tmpfolder):
    # Given options with the namespace extension,
    opts = dict(project_path="my-proj", namespace="ns.ns2", extensions=[Namespace()])

    # when the project is created,
    create_project(opts)

    # then namespace __init__ should not exist
    assert not Path("my-proj/src/ns/__init__.py").exists()
    assert not Path("my-proj/src/ns/ns2/__init__.py").exists()
    # but the package __init__ should
    assert Path("my-proj/src/ns/ns2/my_proj/__init__.py").exists()
    # and plain structure should not exist
    assert not Path("my-proj/src/my_proj/__init__.py").exists()

    # coverage should be configured correctly
    assert "source = ns.ns2.my_proj" in Path("my-proj/.coveragerc").read_text()
    assert "--cov ns.ns2.my_proj" in Path("my-proj/setup.cfg").read_text()


def test_create_project_with_empty_namespace(tmpfolder):
    for j, ns in enumerate(["", None, False]):
        # Given options with the namespace extension,
        opts = dict(project_path=f"my-proj{j}", namespace=ns, extensions=[Namespace()])

        # when the project is created,
        create_project(opts)

        # then plain structure should exist
        path = Path(f"my-proj{j}/src/my_proj{j}/__init__.py")
        assert path.exists()


def test_create_project_without_namespace(tmpfolder):
    # Given options without the namespace extension,
    opts = dict(project_path="proj")

    # when the project is created,
    create_project(opts)

    # then plain structure should exist
    assert Path("proj/src/proj/__init__.py").exists()


def test_cli_with_namespace(tmpfolder):
    # Given the command line with the namespace option,
    sys.argv = ["snek", "proj", "--namespace", "ns"]

    # when snek runs,
    run()

    # then namespace __init__ package should not exist
    assert not Path("proj/src/ns/__init__.py").exists()
    # but the package's should
    assert Path("proj/src/ns/proj/__init__.py").exists()


def test_cli_with_namespace_and_pretend(tmpfolder):
    # Given the command line with the namespace and pretend options
    sys.argv = ["snek", "proj", "--namespace", "ns", "--pretend"]

    # when snek runs,
    run()

    # then namespace __init__ package should not exist (or even the project)
    assert not Path("proj/src/ns/__init__.py").exists()
    assert not Path("proj").exists()


def test_cli_with_empty_namespace(tmpfolder, capsys):
    # Given the command line with the namespace option,
    sys.argv = ["snek", "proj", "--namespace"]

    # when snek runs,
    with pytest.raises(SystemExit):
        run()

    # then an error occurs
    _, err = capsys.readouterr()
    assert "expected one argument" in err


def test_cli_without_namespace(tmpfolder):
    # Given the command line without the namespace option,
    sys.argv = ["snek", "proj"]

    # when snek runs,
    run()

    # then namespace dir should not exist
    assert not Path("proj/src/ns").exists()


def test_move_old_package_without_namespace(tmpfolder):
    # Given a package is already created without namespace
    create_project(project_path="proj", package="my_pkg")

    opts = dict(project_path="proj", package="my_pkg")
    struct = {"src": {"my_pkg": {"file.py": ""}}}

    # when no 'namespace' option is passed,
    struct, opts = get_default_options(struct, opts)
    struct, opts = enforce_namespace_options(struct, opts)
    struct, opts = move_old_package(struct, opts)

    # then the old package remains,
    assert tmpfolder.join("proj/src/my_pkg/__init__.py").check()


def test_move_old_package(tmpfolder):
    # Given a package is already created without namespace
    create_project(project_path="proj", package="my_pkg")
    assert tmpfolder.join("proj/src/my_pkg/__init__.py").check()

    opts = dict(project_path="proj", package="my_pkg", namespace="my.ns")
    struct = {"src": {"my_pkg": {"file.py": ""}}}

    # when the 'namespace' option is passed,
    struct, opts = get_default_options(struct, opts)
    struct, opts = enforce_namespace_options(struct, opts)
    struct, opts = move_old_package(struct, opts)

    # then the old package should be moved
    assert not tmpfolder.join("proj/src/my_pkg/__init__.py").check()
    assert tmpfolder.join("proj/src/my/ns/my_pkg/__init__.py").check()


def test_pretend_move_old_package(tmpfolder, caplog, isolated_logger):
    # Given a package is already created without namespace
    create_project(project_path="proj", package="my_pkg")

    opts = parse_args(["proj", "-p", "my_pkg", "--namespace", "my.ns", "--pretend"])
    struct = {"src": {"my_pkg": {"file.py": ""}}}
    logger.reconfigure(opts)

    # when 'pretend' option is passed,
    struct, opts = get_default_options(struct, opts)
    struct, opts = enforce_namespace_options(struct, opts)
    struct, opts = move_old_package(struct, opts)

    # then nothing should happen,
    assert tmpfolder.join("proj/src/my_pkg/__init__.py").check()
    assert not tmpfolder.join("proj/src/my/ns").check()

    # something should be logged,
    log = caplog.text
    expected_log = ("move", "my_pkg", "to", str(Path("my/ns")))
    for text in expected_log:
        assert text in log

    # but user should see no warning,
    unexpected_warnings = (
        "A folder",
        "exists in the project directory",
        "a namespace option was passed",
        "Please make sure",
    )
    for text in unexpected_warnings:
        assert text not in log


def test_updating_existing_project(tmpfolder, caplog):
    # Given a project already exists, but was generated without
    # namespace,
    create_project(project_path="my-proj")
    assert tmpfolder.join("my-proj/src/my_proj").check()
    assert not tmpfolder.join("my-proj/src/my/ns").check()

    # when the project is updated with a namespace,
    create_project(
        project_path="my-proj", update=True, namespace="my.ns", extensions=[Namespace()]
    )

    # then the package folder should be moved to a nested position,
    assert not tmpfolder.join("my-proj/src/my_proj").check()
    assert tmpfolder.join("my-proj/src/my/ns/my_proj").check()

    # and the user should see a warn
    expected_warnings = (
        "A folder",
        "exists in the project directory",
        "a namespace option was passed",
        "Please make sure",
    )
    log = caplog.text
    for text in expected_warnings:
        assert text in log



import argparse
from shutil import rmtree
from textwrap import dedent

import pytest

from snek import api, cli, info, templates
from snek.extensions import config

from ..helpers import ArgumentParser
from .helpers import make_extension

# ---- "Isolated" tests ----


@pytest.fixture
def default_file(fake_config_dir):
    return fake_config_dir / "default.cfg"


def parse(*args, set_defaults=None):
    parser = ArgumentParser()
    parser.set_defaults(**(set_defaults or {}))
    config.Config().augment_cli(parser)
    return vars(parser.parse_args(args))


def test_no_cli_opts(default_file):
    # When no --config or --save-config is passed
    cli_opts = parse()

    # Then no save_config or config_files should be found in the opts
    assert "save_files" not in cli_opts

    # and config_files will be bootstraped with an empty list
    # if the default file does not exist
    opts = api.bootstrap_options(cli_opts)
    assert opts["config_files"] == []

    # or config_files will be bootstraped with the
    # default file if it exists
    default_file.write_text("[snek]\n")
    opts = api.bootstrap_options(cli_opts)
    assert opts["config_files"] == [default_file]


def test_missing_config():
    # When the --config opt is passed without a value, we have an error
    with pytest.raises((argparse.ArgumentError, TypeError, SystemExit)):
        # ^  TypeError happens because argparse tries to iterate over the --config opts
        #    since it is marked with nargs='+'
        #    ArgumentError and SystemExit might happen depending on the version
        #    of Python when there is a parse error.
        parse("--config")


def test_config_opts(default_file, fake_config_dir):
    files = []
    for j in range(3):
        file = fake_config_dir / f"test{j}.cfg"
        file.write_text("[snek]\n")
        files.append(file)

    # --config can be passed with 1 value
    opts = parse("--config", str(files[0]))
    assert opts["config_files"] == files[:1]

    # --config can be passed with many
    opts = parse("--config", *map(str, files))
    assert opts["config_files"] == files

    # after bootstrap_options, the given files are kept
    # and the default is not included
    opts = api.bootstrap_options(opts)
    assert default_file not in opts["config_files"]
    assert opts["config_files"] == files


def test_no_config():
    opts = parse("--no-config")
    assert opts["config_files"] == api.NO_CONFIG

    # Even after bootstrap_options
    opts = api.bootstrap_options(opts)
    assert opts["config_files"] == api.NO_CONFIG


def test_no_config_conflict(fake_config_dir):
    file = fake_config_dir / "test_no_config.cfg"
    file.write_text("[snek]\n")
    with pytest.raises(SystemExit):
        parse("--no-config", "--config", str(file))


def test_save_config(default_file, fake_config_dir):
    # With no value the default_file is used
    opts = parse("--save-config")
    assert opts["save_config"] == default_file

    # With a value, the value is used
    other_file = fake_config_dir / "other.cfg"
    opts = parse("--save-config", str(other_file))
    assert opts["save_config"] == other_file


# ---- Integration tests ----


def test_save_action(default_file):
    # When the file does not exist
    assert not default_file.exists()
    opts = dict(author="author", email="email", license="MPL-2.0")
    config.save({}, {**opts, "save_config": default_file})
    # it will create a valid file from the point of view of parsing
    assert default_file.exists()
    parsed = info.project({}, default_file)
    assert all(parsed[k] == v for k, v in opts.items())
    # and the file will contain instructions / references
    assert "snek.org" in default_file.read_text()


def test_save_action_no_config(default_file):
    # When the config directory does not exist
    rmtree(str(default_file.parent))
    opts = dict(author="author", email="email", license="MPL-2.0")
    config.save({}, {**opts, "save_config": default_file})
    # it will create the directory automatically when creating the file
    assert default_file.exists()


def existing_config(file):
    text = dedent(
        """\
        [metadata]
        author = John Doe
        author-email = john.joe@fmail.com
        license = gpl3

        [snek]
        # Comment
        version = 3.78
        extensions =
            namespace
            tox
            cirrus
        namespace = my_namespace.my_sub_namespace
        """
    )
    file.write_text(text)
    return file


def test_save_action_existing_file(default_file, monkeypatch):
    # Given default values that differ a bit from the given opts
    monkeypatch.setattr(info, "username", lambda *_: "John Doe")
    monkeypatch.setattr(info, "email", lambda *_: "email")
    monkeypatch.setitem(api.DEFAULT_OPTIONS, "license", "MPL-2.0")
    opts = dict(author="author", email="email", license="MPL-2.0")
    # When the file exists and new config is saved
    existing_config(default_file)
    config.save({}, {**opts, "save_config": default_file})
    # Then metadata that differs from default will change
    expected = dict(author="author", email="john.joe@fmail.com", license="GPL-3.0-only")
    parsed = info.project({}, default_file)
    assert all(parsed[k] == v for k, v in expected.items())


def test_save_action_additional_extensions(default_file):
    # Given the file exists
    existing_config(default_file)
    # When the new config is saved with new extensions
    opts = dict(author="author", email="email", license="MPL-2.0", my_extension1_opt=5)
    extensions = [
        make_extension("MyExtension1"),
        make_extension("MyExtension2"),
        make_extension("MyExtension3", persist=False),
    ]
    config.save({}, {**opts, "save_config": default_file, "extensions": extensions})
    # The old ones are kept and the new ones are added,
    # unless they specify persist=False
    parsed = info.read_setupcfg(default_file).to_dict()["snek"]
    print(default_file.read_text())
    expected = {"namespace", "tox", "cirrus", "my_extension1", "my_extension2"}
    assert templates.parse_extensions(parsed["extensions"]) == expected
    assert parsed["namespace"] == "my_namespace.my_sub_namespace"
    # Extension related opts are also saved
    assert int(parsed["my_extension1_opt"]) == 5


def test_cli_with_save_config(default_file, tmpfolder):
    # Given a global config file does not exist
    assert not default_file.exists()
    # when the CLI is invoked with --save-config
    cli.main("proj -l MPL-2.0 --namespace ns --cirrus --save-config".split())
    # then the file will be created accordingly
    assert default_file.exists()
    parsed = info.read_setupcfg(default_file).to_dict()
    assert parsed["metadata"]["license"] == "MPL-2.0"
    assert parsed["snek"]["namespace"] == "ns"
    assert "cirrus" in parsed["snek"]["extensions"]
    # and since the config extension has persist = False, it will not be stored
    assert "config" not in parsed["snek"]["extensions"]


def test_cli_with_save_config_and_pretend(default_file, tmpfolder):
    # Given a global config file does not exist
    assert not default_file.exists()
    # when the CLI is invoked with --save-config and --pretend
    cli.main("proj --pretend -l MPL-2.0 --namespace ns --cirrus --save-config".split())
    # then the file should not be created
    assert not default_file.exists()
    # (or even the project)
    assert not (tmpfolder / "proj").exists()



#!/usr/bin/env python
import sys
from os.path import exists as path_exists

from snek.api import create_project
from snek.cli import run
from snek.extensions.gitlab_ci import GitLab


def test_create_project_with_gitlab_ci(tmpfolder):
    # Given options with the GitLab extension,
    opts = dict(project_path="proj", extensions=[GitLab()])

    # when the project is created,
    create_project(opts)

    # then files from GitLab extension should exist
    assert path_exists("proj/.gitlab-ci.yml")


def test_create_project_without_gitlab_ci(tmpfolder):
    # Given options without the GitLab extension,
    opts = dict(project_path="proj")

    # when the project is created,
    create_project(opts)

    # then GitLab files should not exist
    assert not path_exists("proj/.gitlab-ci.yml")


def test_cli_with_gitlab_ci(tmpfolder):
    # Given the command line with the GitLab option,
    sys.argv = ["snek", "--gitlab", "proj"]

    # when snek runs,
    run()

    # then files from GitLab and other extensions automatically added should
    # exist
    assert path_exists("proj/.gitlab-ci.yml")
    assert path_exists("proj/tox.ini")
    assert path_exists("proj/.pre-commit-config.yaml")


def test_cli_with_gitlab_ci_and_pretend(tmpfolder):
    # Given the command line with the GitLab and pretend options
    sys.argv = ["snek", "--pretend", "--gitlab", "proj"]

    # when snek runs,
    run()

    # then GitLab files should not exist
    assert not path_exists("proj/.gitlab-ci.yml")
    # (or the project itself)
    assert not path_exists("proj")


def test_cli_without_gitlab_ci(tmpfolder):
    # Given the command line without the GitLab option,
    sys.argv = ["snek", "proj"]

    # when snek runs,
    run()

    # then GitLab files should not exist
    assert not path_exists("proj/.gitlab-ci.yml")



import logging
import os
import re
import stat

from snek import file_system as fs

from .helpers import temp_umask, uniqpath, uniqstr


def test_tmpfile(tmpfolder):
    with fs.tmpfile(suffix=".foo.bar", dir=str(tmpfolder)) as tmp:
        path = str(tmp)
        assert path.endswith(".foo.bar")
        assert tmp.exists()

    assert not tmp.exists()


def test_chdir(caplog, tmpdir, isolated_logger):
    caplog.set_level(logging.INFO)
    curr_dir = os.getcwd()
    dname = uniqstr()  # Use a unique name to get easily identifiable logs
    temp_dir = str(tmpdir.mkdir(dname))
    with fs.chdir(temp_dir):
        new_dir = os.getcwd()
    assert new_dir == os.path.realpath(temp_dir)
    assert curr_dir == os.getcwd()
    assert curr_dir != new_dir
    logs = caplog.text
    assert re.search("chdir.+" + dname, logs)


def test_pretend_chdir(caplog, tmpdir):
    caplog.set_level(logging.INFO)
    curr_dir = os.getcwd()
    dname = uniqstr()  # Use a unique name to get easily identifiable logs
    temp_dir = str(tmpdir.mkdir(dname))
    with fs.chdir(temp_dir, pretend=True):
        new_dir = os.getcwd()
    assert new_dir == curr_dir  # the directory is not changed
    assert curr_dir == os.getcwd()
    logs = caplog.text
    assert re.search("chdir.+" + dname, logs)


def test_create_file(tmpfolder):
    file = fs.create_file("a-file.txt", "content")
    assert file.is_file()
    assert tmpfolder.join("a-file.txt").read() == "content"


def test_pretend_create_file(tmpfolder, caplog):
    caplog.set_level(logging.INFO)
    fname = uniqstr()  # Use a unique name to get easily identifiable logs
    # When a file is created with pretend=True,
    file = fs.create_file(fname, "content", pretend=True)
    # Then it should not be written to the disk,
    assert not file.exists()
    assert tmpfolder.join(fname).check() is False
    # But the operation should be logged
    logs = caplog.text
    assert re.search("create.+" + fname, logs)


def test_create_directory(tmpfolder):
    folder = fs.create_directory("a-dir")
    assert folder.is_dir()
    assert tmpfolder.join("a-dir").check(dir=1)


def test_pretend_create_directory(tmpfolder, caplog):
    caplog.set_level(logging.INFO)
    dname = uniqstr()  # Use a unique name to get easily identifiable logs
    # When a directory is created with pretend=True,
    folder = fs.create_directory(dname, pretend=True)
    # Then it should not appear in the disk,
    assert not folder.exists()
    assert tmpfolder.join(dname).check() is False
    # But the operation should be logged
    logs = caplog.text
    assert re.search("create.+" + dname, logs)


def test_chmod(tmpfolder):
    _ = tmpfolder  # just for chdir

    # Windows have problems with executable bits, so let's just exercise R and W
    with temp_umask(0o333):
        file = uniqpath()
        file.touch()
        assert stat.S_IMODE(file.stat().st_mode) == 0o444
        fs.chmod(file, 0o666)
        assert stat.S_IMODE(file.stat().st_mode) == 0o666


def test_pretend_chmod(tmpfolder, caplog):
    _ = tmpfolder  # just for chdir
    caplog.set_level(logging.INFO)

    # Windows have problems with executable bits, so let's just exercise R and W
    with temp_umask(0o333):
        file = uniqpath()
        file.touch()
        assert stat.S_IMODE(file.stat().st_mode) == 0o444
        # When calling chmod on a file with pretend
        fs.chmod(file, 0o777, pretend=True)
        # It should not change access permissions
        assert stat.S_IMODE(file.stat().st_mode) == 0o444
        # But the operation should be logged
        logs = caplog.text
        assert re.search(f"chmod 777.+{file}", logs)


def test_move(tmpfolder):
    # Given a file or directory exists,
    tmpfolder.join("a-file.txt").write("text")
    tmpfolder.join("a-folder").ensure_dir()
    tmpfolder.join("a-folder/another-file.txt").write("text")
    # When it is moved,
    tmpfolder.join("a-dir").ensure_dir()
    fs.move("a-file.txt", target="a-dir")
    fs.move("a-folder", target="a-dir")
    # Then the original path should not exist
    assert not tmpfolder.join("a-file.txt").check()
    assert not tmpfolder.join("a-folder").check()
    # And the new path should exist
    assert tmpfolder.join("a-dir/a-file.txt").check()
    assert tmpfolder.join("a-dir/a-folder/another-file.txt").check()


def test_move_multiple_args(tmpfolder):
    # Given several files exist,
    tmpfolder.join("a-file.txt").write("text")
    tmpfolder.join("another-file.txt").write("text")
    assert not tmpfolder.join("a-dir/a-file.txt").check()
    assert not tmpfolder.join("a-dir/another-file.txt").check()
    # When they are moved together,
    tmpfolder.join("a-dir").ensure_dir()
    fs.move("a-file.txt", "another-file.txt", target="a-dir")
    # Then the original paths should not exist
    assert not tmpfolder.join("a-file.txt").check()
    assert not tmpfolder.join("another-file.txt").check()
    # And the new paths should exist
    assert tmpfolder.join("a-dir/a-file.txt").read() == "text"
    assert tmpfolder.join("a-dir/another-file.txt").read() == "text"


def test_move_non_dir_target(tmpfolder):
    # Given a file exists,
    tmpfolder.join("a-file.txt").write("text")
    assert not tmpfolder.join("another-file.txt").check()
    # When it is moved,
    fs.move("a-file.txt", target="another-file.txt")
    # Then the original path should not exist
    assert not tmpfolder.join("a-file.txt").check()
    # And the new path should exist
    assert tmpfolder.join("another-file.txt").read() == "text"

    # Given a dir exists,
    tmpfolder.join("a-dir").ensure_dir()
    tmpfolder.join("a-dir/a-file.txt").write("text")
    assert not tmpfolder.join("another-dir/a-file.txt").check()
    # When it is moved to a path that do not exist yet,
    fs.move("a-dir", target="another-dir")
    # Then the dir should be renamed.
    assert not tmpfolder.join("a-dir").check()
    assert tmpfolder.join("another-dir/a-file.txt").read() == "text"


def test_move_log(tmpfolder, caplog):
    caplog.set_level(logging.INFO)
    fname1 = uniqstr()  # Use a unique name to get easily identifiable logs
    fname2 = uniqstr()
    dname = uniqstr()
    # Given a file or directory exists,
    tmpfolder.join(fname1).write("text")
    tmpfolder.join(fname2).write("text")
    fs.move(fname2, target=dname)
    # log should be created.
    logs = caplog.text
    assert re.search(f"move.+{fname2}.+to.+{dname}", logs)


def test_pretend_move(tmpfolder):
    # Given a file or directory exists,
    tmpfolder.join("a-file.txt").write("text")
    tmpfolder.join("another-file.txt").write("text")
    # When it is moved without pretend kwarg,
    tmpfolder.join("a-dir").ensure_dir()
    fs.move("a-file.txt", target="a-dir")
    # Then the src should be moved
    assert tmpfolder.join("a-dir/a-file.txt").check()
    # When it is moved with pretend kwarg,
    fs.move("another-file.txt", target="a-dir", pretend=True)
    # Then the src should not be moved
    assert not tmpfolder.join("a-dir/another-file.txt").check()
    assert tmpfolder.join("another-file.txt").check()


def test_rm_rf(tmp_path):
    # Given nested dirs and files exist
    nested_dir = tmp_path / "dir1/dir2/dir3"
    nested_dir.mkdir(parents=True, exist_ok=True)
    nested_file = nested_dir / "file"
    nested_file.write_text("text")
    # When rm_rf is called
    fs.rm_rf(tmp_path / "dir1")
    # Then nothing should be left
    assert not nested_file.exists()
    assert not nested_dir.exists()
    assert not nested_dir.parent.exists()

    # Given a file exists
    file = tmp_path / "answer.txt"
    file.write_text("42")
    assert file.read_text() == "42"
    # When rm_rf is called with files
    fs.rm_rf(file)
    # Then no exception is raised and the file is removed
    assert not file.exists()


def test_pretend_rm_rf(tmp_path, caplog):
    # Given nested dirs and files exist
    dname = uniqstr()  # Use a unique name to get easily identifiable logs
    nested_dir = tmp_path / dname / "dir2/dir3"
    nested_dir.mkdir(parents=True, exist_ok=True)
    nested_file = nested_dir / "file"
    nested_file.write_text("text")
    # When rm_rf is called with pretend
    fs.rm_rf(tmp_path / dname, pretend=True)
    # Then nothing should change
    assert nested_file.exists()
    assert nested_dir.exists()
    assert nested_dir.parent.exists()
    # But the operation should be logged
    logs = caplog.text
    assert re.search("remove.+" + dname, logs)



import argparse
import builtins
import logging
import os
import os.path
import stat
import traceback
from contextlib import contextmanager
from glob import glob
from pathlib import Path
from pprint import pformat
from shutil import rmtree
from time import sleep
from uuid import uuid4
from warnings import warn

import pytest

skip_on_conda_build = pytest.mark.skipif(
    os.getenv("CONDA_BUILD"),
    reason="conda build does not run inside virtualenv/tox and "
    "it does not have Snek's source code available when running tests",
)


def in_ci() -> bool:
    ci = os.getenv("CI", None) or ""
    return bool(ci.lower() in ("1", "true") or os.getenv("CI_NAME", None))


def uniqstr():
    """Generates a unique random long string every time it is called"""
    return str(uuid4())


def uniqpath(nested=False):
    path = uniqstr()
    if nested:
        path.replace("-", "/")

    return Path(path)


def nop(*args, **kwargs):
    """Function that does nothing"""


def rmpath(path):
    """Carelessly/recursively remove path.
    If an error occurs it will just be ignored, so not suitable for every usage.
    The best is to use this function for paths inside pytest tmp directories, and with
    some hope pytest will also do some cleanup itself.
    """
    try:
        rmtree(str(path), onerror=set_writable)
    except FileNotFoundError:
        return
    except Exception:
        msg = f"rmpath: Impossible to remove {path}, probably an OS issue...\n\n"
        warn(msg + traceback.format_exc())


def path_as_uri(path):
    return Path(os.path.abspath(path)).as_uri()


def set_writable(func, path, exc_info):
    max_attempts = 15
    retry_interval = 0.1
    effective_ids = os.access in os.supports_effective_ids
    existing_files = glob(f"{path}/*")

    if not os.access(path, os.W_OK, effective_ids=effective_ids):
        os.chmod(path, stat.S_IWUSR)
        try:
            return func(path)
        except FileNotFoundError:
            return
    else:
        # For some weird reason we do have rights to remove the dir,
        # let's try again a few times more slowly (maybe a previous OS call
        # returned to the python interpreter but the files were not completely
        # removed yet?)
        for i in range(max_attempts):
            try:
                return rmtree(path)
            except FileNotFoundError:
                return
            except OSError:
                sleep((i + 1) * retry_interval)

    logging.critical(
        "Something went wrong when removing %s. Contents:\n%s",
        path,
        pformat(existing_files),  # weirdly this is usually empty
        exc_info=exc_info,
    )
    (type_, value, traceback) = exc_info
    raise type_(value).with_traceback(traceback)


def command_exception(content):
    # Be lazy to import modules, so coverage has time to setup all the
    # required "probes"
    # (see @FlorianWilhelm comments on #174)
    from snek.exceptions import ShellCommandException

    return ShellCommandException(content)


@contextmanager
def temp_umask(umask):
    """Context manager that temporarily sets the process umask.

    Taken from Python's "private stdlib", module ``test.support``,
    license in https://github.com/python/cpython/blob/master/LICENSE.
    """
    oldmask = os.umask(umask)
    try:
        yield
    finally:
        os.umask(oldmask)


@contextmanager
def disable_import(prefix):
    """Avoid packages being imported

    Args:
        prefix: string at the beginning of the package name
    """
    realimport = builtins.__import__

    def my_import(name, *args, **kwargs):
        if name.startswith(prefix):
            raise ImportError
        return realimport(name, *args, **kwargs)

    try:
        builtins.__import__ = my_import
        yield
    finally:
        builtins.__import__ = realimport


@contextmanager
def replace_import(prefix, new_module):
    """Make import return a fake module

    Args:
        prefix: string at the beginning of the package name
    """
    realimport = builtins.__import__

    def my_import(name, *args, **kwargs):
        if name.startswith(prefix):
            return new_module
        return realimport(name, *args, **kwargs)

    try:
        builtins.__import__ = my_import
        yield
    finally:
        builtins.__import__ = realimport


class ArgumentParser(argparse.ArgumentParser):
    def exit(self, status=0, message=None):
        """Avoid argparse to exit on error"""
        if status:
            raise SystemExit(message, status)



import logging
import os
import re
import subprocess
import sys
from configparser import ConfigParser
from pathlib import Path
from textwrap import dedent
from types import SimpleNamespace as Object

import pytest
from packaging.version import Version

from snek import __path__ as snek_paths
from snek import __version__, actions, info, update
from snek.file_system import chdir

from .helpers import in_ci, path_as_uri, skip_on_conda_build
from .system.helpers import normalize_run_args

EDITABLE_PYSCAFFOLD = re.compile(r"^-e.+snek.*$", re.M | re.I)


class VenvManager:
    def __init__(self, venv):
        self.installed = False
        self.venv = venv
        self.running_version = Version(__version__)
        self._orig_workdir = Path(".").resolve()

    def install(self, pkg=None, editable=False, **kwargs):
        target = [str(pkg)]
        if editable:
            target = ["--editable", *target]

        assert Path(self.venv.exe("python")).exists()
        assert str(self.venv.path) in str(self.venv.exe("python"))

        # Sometimes Windows complain about SSL, despite all the efforts on using a env
        # var for trusted hosts
        cmd = [
            *("pip", "install", *target),
            *("--trusted-host", "files.pythonhosted.org"),
            *("--trusted-host", "pypi.python.org"),
            *("--trusted-host", "files.pythonhosted.org"),
            *("--trusted-host", "pypi.org"),
        ]
        return self.run(*cmd, **kwargs)

    def _get_proj_dir(self) -> Path:
        if "TOXINIDIR" in os.environ:
            # so pytest runs within tox
            proj_dir = Path(os.environ["TOXINIDIR"])
            logging.debug("SRC via TOXINIDIR: %s", proj_dir)
        else:
            try:
                location = Path(snek_paths[0])
                assert location.parent.name == "src"
                proj_dir = location.parent.parent
            except:  # noqa
                print("\n\nInstall Snek with python setup.py develop!\n\n")
                raise

            logging.debug("SRC via working_set: %s, location: %s", proj_dir, location)

        return proj_dir

    def _install_pre_built_wheel(self, proj_dir: Path):
        # CI should pre-build wheels that can be used.
        candidates = (proj_dir / "dist").glob("Snek*.whl")
        wheel = next(iter(sorted(candidates, reverse=True, key=str)), None)
        assert wheel, "Snek should be pre-built by CI, but it is not..."
        return self.install(path_as_uri(wheel))

    def _install_from_src(self, proj_dir: Path):
        env = {**os.environ, "SETUPTOOLS_SCM_PRETEND_VERSION": __version__}
        assert proj_dir.exists(), f"{proj_dir} is supposed to exist"
        out = self.install(proj_dir, editable=True, env=env)
        pkg_list = self.run("pip freeze")
        assert EDITABLE_PYSCAFFOLD.findall(pkg_list)
        return out

    def install_this_snek(self):
        # Normally the following command should do the trick
        # self.venv.install_package('Snek')
        # but sadly pytest-virtualenv chokes on the src-layout of Snek
        proj_dir = self._get_proj_dir()
        if in_ci():
            self._install_pre_built_wheel(proj_dir)
        else:
            self._install_from_src(proj_dir)

        # Make sure snek was not installed using PyPI
        assert self.running_version.public <= self.snek_version().public
        self.installed = True
        return self

    def install_snek(self, major, minor):
        ver = f"snek>={major}.{minor},<{major}.{minor + 1}a0"
        self.install(ver)
        installed_version = self.snek_version().release[:2]
        assert installed_version == (major, minor)
        self.installed = True
        return self

    def uninstall_snek(self):
        self.run("pip uninstall -y snek")
        putup = self.venv.exe("putup")
        assert str(self.venv.path.resolve()) not in str(Path(putup).resolve())
        self.installed = False
        return self

    def snek_version(self):
        try:
            cli_version = self.run("python -m snek.cli --version").lower()
            return Version(cli_version.replace("snek ", ""))
        except subprocess.CalledProcessError as ex:
            if (
                sys.version_info >= (3, 12)
                and "No module named 'snek.contrib.six.moves'" in ex.output
            ):
                pytest.skip("Cannot import from six.moves in Python >= 3.12")
            raise

    def putup(self, *args, **kwargs):
        args, kwargs = normalize_run_args(args, kwargs)
        print("putup", *args)
        print(self.venv.run("putup", *args, **kwargs))
        return self

    def run(self, *args, **kwargs):
        return self.venv.run(*args, **kwargs)


@pytest.fixture
def venv_mgr(venv):
    return VenvManager(venv)


@pytest.mark.slow
@pytest.mark.requires_src
@skip_on_conda_build
def test_update_version_3_0_to_3_1(tmp_path, with_coverage, venv_mgr):
    with chdir(str(tmp_path)):
        name = "my_old_project"
        project = tmp_path / "my_old_project"
        (
            venv_mgr.install_snek(3, 0)
            .putup(name)
            .uninstall_snek()
            .install_this_snek()
            .putup(f"--update {project}", with_coverage=with_coverage)
        )
    setup_cfg = Path(project, "setup.cfg").read_text(encoding="utf-8")
    assert "[options.entry_points]" in setup_cfg


@pytest.mark.slow
@pytest.mark.requires_src
@skip_on_conda_build
def test_update_version_3_0_to_3_1_pretend(tmp_path, with_coverage, venv_mgr):
    with chdir(str(tmp_path)):
        name = "my_old_project"
        project = tmp_path / "my_old_project"
        (
            venv_mgr.install_snek(3, 0)
            .putup(name)
            .uninstall_snek()
            .install_this_snek()
            .putup(f"--pretend --update {project}", with_coverage=with_coverage)
        )
    setup_cfg = Path(project, "setup.cfg").read_text(encoding="utf-8")
    assert "[options.entry_points]" not in setup_cfg


@pytest.mark.slow
@pytest.mark.requires_src
@skip_on_conda_build
def test_inplace_update(tmp_path, with_coverage, venv_mgr):
    # Given an existing project
    project = tmp_path / "my_old_project"
    cmd = f"--name my-ns-proj --package project --namespace my_ns {project}"
    venv_mgr.install_this_snek().putup(cmd)

    # With an existing configuration
    assert project / "setup.cfg"
    parser = ConfigParser()
    parser.read(project / "setup.cfg")
    assert parser["metadata"]["name"] == "my-ns-proj"
    assert parser["snek"]["package"] == "project"
    assert parser["snek"]["namespace"] == "my_ns"

    # And without some extensions
    for file in (".pre-commit-config.yaml", ".isort.cfg"):
        assert not Path(project, file).exists()

    # When the project is updated
    # without repeating the information already given
    # but adding some information/extensions
    with chdir(str(project)):
        venv_mgr.putup(
            "-vv --description asdf --pre-commit --update .",
            with_coverage=with_coverage,
        )

    # Then existing configuration should be preserved + the additions
    parser = ConfigParser()
    parser.read(project / "setup.cfg")
    assert parser["metadata"]["name"] == "my-ns-proj"
    assert parser["snek"]["package"] == "project"
    assert parser["snek"]["namespace"] == "my_ns"

    # Some information (metadata) require manual update
    # unless the --force option is used
    assert parser["metadata"]["description"] != "asdf"

    # New extensions should take effect
    for file in ("tox.ini", ".pre-commit-config.yaml", ".isort.cfg"):
        assert Path(project, file).exists()

    # While using the existing information
    parser = ConfigParser()
    parser.read(project / ".isort.cfg")
    assert parser["settings"]["known_first_party"] == "my_ns"


# ---- Slightly more isolated tests ----


def test_update_setup_cfg(tmpfolder):
    # Given an existing setup.cfg
    proj = Path(tmpfolder, "proj")
    proj.mkdir(parents=True, exist_ok=True)
    (proj / "setup.cfg").write_text("[metadata]\n\n[snek]\n")
    # when we update it
    extensions = [Object(name="cirrus", persist=True), Object(name="no", persist=False)]
    opts = {"project_path": proj, "extensions": extensions}
    _, opts = actions.get_default_options({}, opts)
    update.update_setup_cfg({}, opts)
    cfg = info.read_setupcfg(proj / "setup.cfg")
    # then it should show the most update snek version
    assert cfg["snek"]["version"].value == __version__
    assert "cirrus" in cfg["snek"]["extensions"].value
    assert "no" not in cfg["snek"]["extensions"].value
    # and some configuration keys should be present
    assert "options" in cfg


def test_update_none_param(tmpfolder):
    invalid = """\
    [metadata]
    [snek]
    version = 4
    """
    Path(tmpfolder, "setup.cfg").write_text(dedent(invalid))
    extensions = [Object(name="x_foo_bar_x", persist=True)]
    _, opts = actions.get_default_options({}, {"extensions": extensions})
    opts["x_foo_bar_x_param"] = None
    # No parser exception should be found
    update.update_setup_cfg({}, opts)
    assert Path(tmpfolder, "setup.cfg").read_text()


def test_add_dependencies(tmpfolder):
    # Given an existing setup.cfg
    Path(tmpfolder, "setup.cfg").write_text("[options]\n")
    # when we update it
    opts = {"project_path": tmpfolder, "pretend": False}
    update.add_dependencies({}, opts)
    # then we should see the dependencies in install_requires
    cfg = info.read_setupcfg(Path(tmpfolder, "setup.cfg"))
    assert "install_requires" in str(cfg["options"])
    assert "importlib-metadata" in str(cfg["options"]["install_requires"])


def test_add_dependencies_with_comments(tmpfolder):
    # Given a setup.cfg with comments inside options (especially install_requires)
    config = """\
    [metadata]
    project_urls =
        Download = https://pypi.org/project/Snek/#files
        # the previous line does not have a comment, it's just part of the URL
        Issues = https://github.com/snek/snek/issues  # this is a comment!
    [options]
    install_requires =
        importlib-metadata; python_version<"3.8"
        # Adding some comments here that are perfectly valid.
        some-other-dependency
        gitdep @ git+https://repo.com/gitdep@main#egg=gitdep
    packages = find_namespace:
    """
    Path(tmpfolder, "setup.cfg").write_text(dedent(config))
    # when we update it
    opts = {"project_path": tmpfolder, "pretend": False}
    update.add_dependencies({}, opts)
    # then the comments inside the options should still be there
    actual_setup_cfg = Path(tmpfolder, "setup.cfg").read_text()
    assert actual_setup_cfg == dedent(config)


@pytest.fixture
def existing_config(tmpfolder):
    config = """\
    [options]
    setup_requires =
        snek
        somedep>=3.8

    [snek]
    version = 3.2.2
    """
    cfg = Path(tmpfolder) / "setup.cfg"
    cfg.write_text(dedent(config))

    yield cfg


def test_handover_setup_requires(tmpfolder, existing_config):
    # Given an existing setup.cfg with setup_requires
    # when we update it
    opts = {"project_path": tmpfolder, "pretend": False}
    update.handover_setup_requires({}, opts)
    cfg = info.read_setupcfg(existing_config)
    # then setup_requirements should not be included
    assert "setup_requires" not in str(cfg["options"])


def test_handover_setup_requires_no_pyproject(tmpfolder, existing_config):
    # Given an existing setup.cfg with outdated setup_requires and snek version,
    # when we update it without no_pyproject
    opts = {"project_path": tmpfolder, "pretend": False, "isolated_build": False}
    update.handover_setup_requires({}, opts)
    cfg = info.read_setupcfg(existing_config)
    # then setup_requirements is left alone
    assert cfg["options"]["setup_requires"]


@pytest.fixture
def pyproject_from_old_extension(tmpfolder):
    """Old pyproject.toml file as produced by snekext-pyproject"""
    config = """\
    [build-system]
    requires = ["setuptools"]
    """
    pyproject = Path(tmpfolder) / "pyproject.toml"
    pyproject.write_text(dedent(config))
    yield pyproject


def test_update_pyproject_toml(tmpfolder, pyproject_from_old_extension):
    update.update_pyproject_toml({}, {"project_path": tmpfolder, "pretend": False})
    pyproject = info.read_pyproject(pyproject_from_old_extension)
    deps = " ".join(pyproject["build-system"]["requires"])
    assert "setuptools_scm" in deps
    assert "setuptools.build_meta" in pyproject["build-system"]["build-backend"]
    assert "setuptools_scm" in pyproject["tool"]


def test_migrate_setup_requires(tmpfolder, existing_config):
    # When a project with setup.cfg :: setup_requires is updated
    opts = {"project_path": tmpfolder, "pretend": False}
    _, opts = update.handover_setup_requires({}, opts)
    update.update_pyproject_toml({}, opts)
    # then the minimal dependencies are added
    pyproject = info.read_pyproject(tmpfolder)
    deps = " ".join(pyproject["build-system"]["requires"])
    assert "setuptools_scm" in deps
    # old dependencies are migrated from setup.cfg
    assert "somedep>=3.8" in deps
    setupcfg = info.read_setupcfg(existing_config)
    assert "setup_requires" not in setupcfg["options"]
    # but snek is not included.
    assert "snek" not in deps


def test_replace_find_with_find_namespace(tmpfolder):
    # Given an old setup.cfg based on packages find:
    config = """\
    [options]
    zip_safe = False
    packages = find:

    [options.packages.find]
    where = src
    exclude =
        tests
    """
    Path(tmpfolder, "setup.cfg").write_text(dedent(config))
    # when we update it
    opts = {"project_path": tmpfolder, "pretend": False}
    update.replace_find_with_find_namespace({}, opts)
    # then we should see find_namespace instead
    cfg = info.read_setupcfg(Path(tmpfolder, "setup.cfg"))
    assert cfg["options"]["packages"].value == "find_namespace:"
    assert "options.packages.find" in cfg
    assert cfg["options.packages.find"]["where"].value == "src"
    assert cfg["options.packages.find"]["exclude"].value.strip() == "tests"



from collections.abc import MutableMapping
from textwrap import dedent

from snek import toml


def test_dumps():
    serialised = toml.dumps({"test": {"key": 42}})
    assert isinstance(serialised, str)
    assert "[test]" in serialised
    assert "key = 42" in serialised

    unserialised = toml.loads(serialised)
    assert "test" in unserialised
    assert "key" in unserialised["test"]
    assert unserialised["test"]["key"] == 42


def test_loads():
    example = """\
    [test]
    key = 42
    list = [1, 2, 3, 4]

    [nested.table.nested]
    x = "y"
    """

    unserialised = toml.loads(dedent(example))

    assert isinstance(unserialised, MutableMapping)
    assert "test" in unserialised
    assert "key" in unserialised["test"]
    assert unserialised["test"]["key"] == 42
    assert unserialised["nested"]["table"]["nested"]["x"] == "y"
    assert unserialised["test"]["list"] == [1, 2, 3, 4]

    serialised = toml.dumps(unserialised)
    assert "[test]" in serialised
    assert "key = 42" in serialised


def test_setdefault():
    x = {}
    toml.setdefault(x, "a.b.c.d", 42)
    assert x["a"]["b"]["c"]["d"] == 42

    y = toml.loads(toml.dumps({}))
    assert "a" not in y
    toml.setdefault(y, "a.b.c.d", 42)
    assert y["a"]["b"]["c"]["d"] == 42

    # When key is already set, setdefault returns the existing value
    # and do not change the object
    assert toml.setdefault(y, "a.b.c.d", -1) == 42
    assert y["a"]["b"]["c"]["d"] == 42



"""
Functionality for working with a git repository
"""

from pathlib import Path
from typing import Optional, TypeVar, Union

from . import shell
from .exceptions import ShellCommandException
from .file_system import PathLike, chdir
from .log import logger

T = TypeVar("T")


def git_tree_add(struct: dict, prefix: PathLike = "", **kwargs):
    """Adds recursively a directory structure to git

    Args:
        struct: directory structure as dictionary of dictionaries
        prefix: prefix for the given directory structure

    Additional keyword arguments are passed to the
    :obj:`git <snek.shell.ShellCommand>` callable object.
    """
    prefix = Path(prefix)
    for name, content in struct.items():
        if isinstance(content, dict):
            git_tree_add(struct[name], prefix=prefix / name, **kwargs)
        elif content is None or isinstance(content, str):
            shell.git("add", str(prefix / name), **kwargs)
        else:
            raise TypeError(f"Don't know what to do with content type {type}.")


def add_tag(project: PathLike, tag_name: str, message: Optional[str] = None, **kwargs):
    """Add an (annotated) tag to the git repository.

    Args:
        project: path to the project
        tag_name: name of the tag
        message: optional tag message

    Additional keyword arguments are passed to the
    :obj:`git <snek.shell.ShellCommand>` callable object.
    """
    with chdir(project):
        if message is None:
            shell.git("tag", tag_name, **kwargs)
        else:
            shell.git("tag", "-a", tag_name, "-m", message, **kwargs)


def init_commit_repo(project: PathLike, struct: dict, **kwargs):
    """Initialize a git repository

    Args:
        project: path to the project
        struct: directory structure as dictionary of dictionaries

    Additional keyword arguments are passed to the
    :obj:`git <snek.shell.ShellCommand>` callable object.
    """
    logger.report("initialize", f"git repo in {project}...")
    with chdir(project, pretend=kwargs.get("pretend")):
        shell.git("init", **kwargs)
        git_tree_add(struct, **kwargs)
        shell.git("commit", "-m", "Initial commit", **kwargs)


def is_git_repo(path: PathLike):
    """Check if path is a git repository"""
    path = Path(path)
    logger.report("check", f"is {path} already a git repo...")
    if not path.is_dir():
        return False

    with chdir(path):
        try:
            shell.git("rev-parse", "--git-dir")
        except ShellCommandException:
            return False
        return True


def get_git_root(default: Optional[T] = None) -> Union[None, T, str]:
    """Return the path to the top-level of the git repository or *default*.

    Args:
        default: if no git root is found, default is returned

    Returns:
        str: top-level path or *default*
    """
    try:
        return next(shell.git("rev-parse", "--show-toplevel"))
    except ShellCommandException:
        return default



import os
import sys
from pathlib import Path
from subprocess import CalledProcessError

import pytest

from snek.file_system import chdir
from snek.info import read_pyproject

from ..helpers import skip_on_conda_build
from .helpers import run, run_common_tasks

pytestmark = [pytest.mark.slow, pytest.mark.system]


def is_venv():
    """Check if the tests are running inside a venv"""
    return hasattr(sys, "real_prefix") or (
        hasattr(sys, "base_prefix") and sys.base_prefix != sys.prefix
    )


@pytest.fixture(autouse=True)
def cwd(tmpdir):
    """Guarantee a blank folder as workspace"""
    with tmpdir.as_cwd():
        yield tmpdir


@skip_on_conda_build
@pytest.mark.requires_src
def test_ensure_inside_test_venv(putup):
    # This is a METATEST
    # Here we ensure `putup` is installed inside tox so we know we are testing the
    # correct version of snek and not one the devs installed to use in other
    # projects
    assert ".tox" in putup


BUILD_DEPS = ["setuptools_scm"]


def test_putup(cwd, putup):
    # Given snek is installed,
    # when we run putup
    run(f"{putup} myproj")
    # then no error should be raised when running the common tasks
    with cwd.join("myproj").as_cwd():
        # then the new version of Snek should produce packages with
        # the correct build deps
        pyproject_toml = read_pyproject(".")
        stored_deps = " ".join(pyproject_toml["build-system"]["requires"])
        for dep in BUILD_DEPS:
            assert dep in stored_deps
        # and no error should be raised when running the common tasks
        run_common_tasks()


def test_putup_with_update(cwd, putup):
    # Given snek is installed,
    # and a project already created
    run(f"{putup} myproj")
    # when we run putup with the update flag
    run(f"{putup} --update myproj")
    # then no difference should be found
    with cwd.join("myproj").as_cwd():
        git_diff = run("git diff")
        assert git_diff.strip() == ""


def test_putup_with_update_dirty_workspace(cwd, putup):
    run(f"{putup} myproj")
    with chdir("myproj"):
        with open("setup.py", "w") as fh:
            fh.write("DIRTY")
    with pytest.raises(CalledProcessError):
        run(f"{putup} --update myproj")
    run(f"{putup} --update myproj --force")


def test_putup_with_update_and_namespace(cwd, putup):
    # Given snek is installed,
    # and a project already created
    run(f"{putup} --namespace myns --package myproj myns-myproj")
    # when we run putup with the update flag
    run(f"{putup} --update myns-myproj")
    # then no difference should be found
    with cwd.join("myns-myproj").as_cwd():
        git_diff = run("git diff")
        assert git_diff.strip() == ""


def test_differing_package_name(cwd, putup):
    # Given snek is installed,
    # when we run putup
    run(f"{putup} my-cool-proj -p myproj")
    # then the folder structure should respect the names
    assert Path("my-cool-proj").is_dir()
    assert Path("my-cool-proj/src/myproj").is_dir()
    # then no error should be raised when running the common tasks
    with cwd.join("my-cool-proj").as_cwd():
        run_common_tasks()


def test_update(putup):
    # Given snek is installed,
    # and a project already created
    run(f"{putup} myproj")
    assert not Path("myproj/.cirrus.yml").exists()
    # when it is updated
    run(f"{putup} --update --cirrus myproj")
    # then complementary files should be created
    assert Path("myproj/.cirrus.yml").exists()


def test_force(cwd, putup):
    # Given snek is installed,
    # and a project already created
    run(f"{putup} myproj")
    assert not Path("myproj/.cirrus.yml").exists()
    # when it is forcefully updated
    run(f"{putup} --force --cirrus myproj")
    # then complementary files should be created
    assert Path("myproj/.cirrus.yml").exists()


# -- Extensions --


def test_tox_docs(cwd, tox, putup):
    # Given snek project is created
    run(f"{putup} myproj")
    with cwd.join("myproj").as_cwd():
        # when we can call tox -e docs
        run(f"{tox} -e docs")
        # then documentation will be generated.
        assert Path("docs/api/modules.rst").exists()
        assert Path("docs/_build/html/index.html").exists()


def test_tox_doctests(cwd, tox, putup):
    # Given snek project is created
    run(f"{putup} myproj")
    with cwd.join("myproj").as_cwd():
        # when we can call tox
        run(f"{tox} -e doctests")
        # then tests will execute


def test_tox_tests(cwd, tox, putup):
    # Given snek project is created
    run(f"{putup} myproj")
    with cwd.join("myproj").as_cwd():
        # when we can call tox
        run(tox)
        # then tests will execute


def test_tox_build(cwd, tox, putup):
    # Given snek project is created
    run(f"{putup} myproj")
    with cwd.join("myproj").as_cwd():
        # when we can call tox
        # then tasks will execute
        run(f"{tox} -e build")
        # and a pure Python distribution is created
        assert len(list(Path("dist").glob("*py3-none-any.whl"))) > 0
        try:
            run(f"{tox} -e clean")
            assert not Path("dist").exists()
            assert not Path("build").exists()
        except CalledProcessError as ex:
            msg = (ex.stdout or "") + (ex.stderr or "")
            if os.name == "nt" and ("unicodeescape" in msg):
                pytest.skip("Sometimes Windows have problems with rmtree")


@pytest.mark.parametrize(
    "extension, kwargs, filename",
    (
        ("pre-commit", {}, ".pre-commit-config.yaml"),
        ("cirrus", {}, ".cirrus.yml"),
        ("gitlab", {}, ".gitlab-ci.yml"),
    ),
)
def test_extensions(cwd, putup, extension, kwargs, filename):
    # Given snek is installed,
    # when we call putup with extensions
    name = "myproj-" + extension
    run(f"{putup} -vv --{extension} {name}")
    with cwd.join(name).as_cwd():
        # then special files should be created
        assert Path(filename).exists()
        # and all the common tasks should run properly
        run_common_tasks(**kwargs)


def test_no_cli(cwd, putup):
    # Given snek is installed,
    # when we call putup with --no-cli
    run(f"{putup} myproj --no-cli")
    with cwd.join("myproj").as_cwd():
        # then no cli file should be created
        assert not Path("src/myproj/cli.py").exists()
        assert not Path("tests/test_cli.py").exists()
        # and all the common tasks should run properly
        run_common_tasks(tests=False)


def test_namespace(cwd, putup):
    # Given snek is installed,
    # when we call putup with --namespace
    run(f"{putup} nested_project -p my_package --namespace com.blue_yonder")
    # then a very complicated module hierarchy should exist
    path = "nested_project/src/com/blue_yonder/my_package/cli.py"
    assert Path(path).exists()
    assert not Path("nested_project/src/my_package").exists()
    with cwd.join("nested_project").as_cwd():
        run_common_tasks()
        # sphinx should be able to document modules that use PEP 420
        assert Path("docs/api/com.blue_yonder.my_package.rst").exists()
    # and snek should remember the options during an update
    run(f"{putup} nested_project --update -vv")
    assert Path(path).exists()
    assert not Path("nested_project/src/nested_project").exists()
    assert not Path("nested_project/src/my_package").exists()


def test_namespace_no_cli(cwd, putup):
    # Given snek is installed,
    # when we call putup with --namespace and --no-cli
    run(
        f"{putup} nested_project --no-cli "
        "-p my_package --namespace com.blue_yonder"
    )
    # then a very complicated module hierarchy should exist
    path = Path("nested_project/src/com/blue_yonder/my_package")
    assert path.is_dir()
    # but no cli.py
    assert not (path / "cli.py").exists()


@pytest.mark.skipif(sys.version_info < (3, 7), reason="pre-commit requires Python 3.7+")
def test_new_project_does_not_fail_pre_commit(cwd, pre_commit, putup):
    # Given snek is installed,
    # when we call putup with extensions and pre-commit
    name = "my_project"
    run(
        f"{putup} --pre-commit --cirrus --gitlab "
        "-p my_package --namespace com.blue_yonder " + name
    )
    with cwd.join(name).as_cwd():
        # then the newly generated files should not result in errors when
        # pre-commit runs...
        try:
            run(f"{pre_commit} install")
            run(f"{pre_commit} run --all")
        except CalledProcessError as ex:
            if os.name == "nt" and (
                "filename or extension is too long"
                in ((ex.stdout or "") + (ex.stderr or ""))
            ):
                pytest.skip("Sometimes Windows have problems with nested files")
                # Even if we try to change that by configuring the CI
                # environment
            else:
                raise



#!/usr/bin/env python
import sys
from os.path import exists as path_exists

from snek.api import create_project
from snek.cli import run
from snek.extensions.cirrus import Cirrus


def test_create_project_with_cirrus(tmpfolder):
    # Given options with the cirrus extension,
    opts = dict(project_path="proj", extensions=[Cirrus("cirrus")])

    # when the project is created,
    create_project(opts)

    # then files from cirrus extension should exist
    assert path_exists("proj/.cirrus.yml")


def test_create_project_without_cirrus(tmpfolder):
    # Given options without the cirrus extension,
    opts = dict(project_path="proj")

    # when the project is created,
    create_project(opts)

    # then cirrus files should not exist
    assert not path_exists("proj/.cirrus.yml")


def test_cli_with_cirrus(tmpfolder):
    # Given the command line with the cirrus option,
    sys.argv = ["snek", "--cirrus", "proj"]

    # when snek runs,
    run()

    # then files from cirrus and other extensions automatically added should
    # exist
    assert path_exists("proj/.cirrus.yml")
    assert path_exists("proj/tox.ini")
    assert path_exists("proj/.pre-commit-config.yaml")


def test_cli_with_cirrus_and_pretend(tmpfolder):
    # Given the command line with the cirrus and pretend options
    sys.argv = ["snek", "--pretend", "--cirrus", "proj"]

    # when snek runs,
    run()

    # then cirrus files should not exist
    assert not path_exists("proj/.cirrus.yml")
    # (or the project itself)
    assert not path_exists("proj")


def test_cli_without_cirrus(tmpfolder):
    # Given the command line without the cirrus option,
    sys.argv = ["snek", "proj"]

    # when snek runs,
    run()

    # then cirrus files should not exist
    assert not path_exists("proj/.cirrus.yml")
